# -*- python -*-
# $RCSfile: pixelinfodisplay.py,v $
# $Revision: 1.12 $
# $Author: langer $
# $Date: 2006/04/21 17:52:12 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

from oof2.common import color
from oof2.common import debug
from oof2.common import primitives
from oof2.common import registeredclass
from oof2.common.IO import display
from oof2.common.IO import parameter
from oof2.common.IO import xmlmenudump

class PixelInfoDisplay(display.DisplayMethod):
    def __init__(self, color, line_width):
        self.color = color
        self.line_width = line_width
        display.DisplayMethod.__init__(self)

    def draw(self, gfxwindow, device):
        toolbox = gfxwindow.getToolboxByName("Pixel_Info")
        pixel = toolbox.currentPixel()
        if pixel is not None:
            microstructure = toolbox.findMicrostructure()
            if microstructure is not None:
                self.drawPixel(device, pixel, microstructure)

#      n3_______________n2 ((i+1)*dx, (j+1)*dy)
#       |               |
#       |               |
#       |               |
#       |               dy
#       |     (i, j)    |
#       |               |
#       |               |
#       |               |
#       n0------dx------n1
# (i*dx, j*dy)
    def drawPixel(self, device, pixel, microstructure):
        dx = microstructure.sizeOfPixels()[0]
        dy = microstructure.sizeOfPixels()[1]
        i = pixel.x
        j = pixel.y
        n0 = primitives.Point(i*dx, j*dy)
        n1 = primitives.Point((i+1)*dx, j*dy)
        n2 = primitives.Point((i+1)*dx, (j+1)*dy)
        n3 = primitives.Point(i*dx, (j+1)*dy)
        device.set_lineColor(self.color)
        device.set_lineWidth(self.line_width)
        device.draw_segment(primitives.Segment(n0, n1))
        device.draw_segment(primitives.Segment(n1, n2))
        device.draw_segment(primitives.Segment(n2, n3))
        device.draw_segment(primitives.Segment(n3, n0))

    def getTimeStamp(self, gfxwindow):
        toolbox = gfxwindow.getToolboxByName("Pixel_Info")
        return max(self.timestamp,
                   toolbox.getTimeStamp())

pixelInfoDisplay = registeredclass.Registration(
    'Pixel Info',
    display.DisplayMethod,
    PixelInfoDisplay,
    params=[color.ColorParameter('color', color.blue,
                                 tip="Color for the queried pixel."),
            parameter.IntRangeParameter('line_width', (0,10), 0,
                                        tip="Line width.")],
    ordering=2.70,
    layerordering=display.PointLike(0.5),
    whoclasses=('Image','Microstructure', 'Top Bitmap'),
    tip="Display a pixel that is being queried.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/common/reg/pixelinfodisplay.xml')
    )

########################

# In order to have a single layer that displays pixel info for *both*
# Images and Microstructures, we create a new WhoClass that contains
# only a single WhoProxy object.  The proxy resolves to the topmost
# Image or Microstructure.  We can then create a predefined layer for
# the WhoClass.

from oof2.common.IO import whoville

class TopBitmap(whoville.WhoProxyClass):
    def resolve(self, proxy, gfxwindow):
        return gfxwindow.topwho('Microstructure', 'Image')
    def getTimeStamp(self, proxy, gfxwindow):
        return gfxwindow.getLayerChangeTimeStamp()

TopBitmap('<top bitmap>')

whoville.WhoClass('Top Bitmap', ordering=20000, proxyClasses=['<top bitmap>'])

display.predefineLayer('Top Bitmap', '<top bitmap>', pixelInfoDisplay())
