# -*- python -*-
# $RCSfile: outputClones.py,v $
# $Revision: 1.69 $
# $Author: reida $
# $Date: 2006/06/28 18:53:33 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

# A smattering of pre-defined outputs for the convenience of OOF users.

from oof2.SWIG.common import coord
from oof2.SWIG.common import ooferror
from oof2.SWIG.engine import invariant
from oof2.SWIG.engine import outputval
from oof2.common import debug
from oof2.common import enum
from oof2.common import primitives
from oof2.common import utils
from oof2.common.IO import output
from oof2.common.IO import parameter
from oof2.engine.IO import meshparameters
from types import *
import math

import sys

# Examples of Outputs

# See comments in output.py.

#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#

def _pos(mesh, elements, coords):
    # The argument "elements" is a list of Elements or Edges, possibly
    # mixed together. The argument "coords" is a list of lists.  Each
    # Element in the elements list corresponds to a list of
    # MasterCoords in the coords list, and each Edge corresponds to a
    # list of doubles (ie, master coords for the Edge, in the range
    # [0,1)).
    return utils.flatten1([element.position(coordList)
                           for element,coordList in zip(elements, coords)])

posOutput = output.Output(
    name="original position",
    otype=(primitives.Point, coord.Coord),
    callback=_pos)
                            

#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#

class SpaceComponent(enum.EnumClass('x', 'y', 'z')):
    tip='Components of vectors.'
    discussion="""<para>
    <classname>SpaceComponent</classname> is used by various
    <link
    linkend='Section:Output'><classname>Outputs</classname></link> to
    choose one of the components of a vector quantity.
    </para>"""
    
class InPlaneSpaceComponent(enum.EnumClass('x', 'y')):
    tip="The in-plane components of vectors."
    discussion="""<para>
    <classname>InPlaneSpaceComponent</classname> is used by various
    <link
    linkend='Section:Output'><classname>Outputs</classname></link> to
    choose one of the in-plane components of a vector quantity.
    </para>"""

#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#

# FieldOutput is for fields which are defined on the Mesh.
# FieldAnywayOutput is for fields, even if they're not defined on the
# mesh.

def _field(mesh, elements, coords, field):
    return utils.flatten1([element.outputFields(mesh, field, ecoords)
           for element, ecoords in zip(elements, coords)])

FieldOutput = output.Output(
    name = "field",
    callback = _field,
    otype = outputval.OutputValPtr,
    params = [meshparameters.FieldParameter("field", outofplane=1,
                                               tip=parameter.emptyTipString)],
    tip="Compute Field values.",
    discussion='<para>Compute the value of the given &field; on a &mesh;.</para>'
    )

# The FieldAnywayOutput is just like a FieldOutput, but it is
# computable even if the given Field isn't defined on the Mesh.  The
# Field values are assumed to be zero if the field isn't defined.
# This is useful especially for Displacement and positionOutputs.

def _fieldAnyway(mesh, elements, coords, field):
    return utils.flatten1([element.outputFieldsAnyway(mesh, field, ecoords)
                           for element, ecoords in zip(elements, coords)])

FieldAnywayOutput = output.Output(
    name = "field anyway",
    callback=_fieldAnyway,
    otype=outputval.OutputValPtr,
    params = [meshparameters.FieldAnywayParameter("field", outofplane=1)])

############

def _fieldderiv(mesh, elements, coords, field, derivative):
    return utils.flatten1(
        [element.outputFieldDerivs(mesh, field, derivative, ecoords)
         for element, ecoords in zip(elements, coords)])

FieldDerivOutput = output.Output(
    name = "field derivative",
    callback = _fieldderiv,
    otype = outputval.OutputValPtr,
    params = [meshparameters.FieldParameter("field", outofplane=1,
                                                tip=parameter.emptyTipString),
                  enum.EnumParameter("derivative", InPlaneSpaceComponent,
                                     tip='Which derivative to take.')],
    tip='Compute derivatives of Fields.',
    discussion='<para>Compute the spatial derivative of a &field; on a &mesh;.</para>')

############

def _flux(mesh, elements, coords, flux):
    ans = []
    for element, ecoords in zip(elements, coords):
        element.begin_material_computation(mesh)
        ans.append(element.outputFluxes(mesh, flux, ecoords))
        element.end_material_computation(mesh)
    return utils.flatten1(ans)

FluxOutput = output.Output(
    name = "flux",
    callback = _flux,
    otype = outputval.OutputValPtr,
    params = [meshparameters.FluxParameter("flux",
                                              tip=parameter.emptyTipString)],
    tip='Compute Flux values.',
    discussion='<para>Compute the value of the given &flux; on a &mesh;.</para>')

################

def _component(mesh, elements, coords, field, component):
    if field:
        # 'component' is a string, "x" or "xy" or the like
        comp = field[0].getIndex(component) # convert string to IndexP
        return [f[comp] for f in field]
    return []
        
ComponentOutput = output.Output(
    name = "component",
    callback = _component,
    otype = FloatType,
    inputs = [outputval.OutputValParameter('field')],
    params = [meshparameters.FieldIndexParameter('component',
                                                 tip='Which component to take.')
              ]
    )

FieldCompOutput = ComponentOutput.clone(
    name="field component",
    tip='Compute a component of a Field.',
    discussion=
    '<para>Compute a single component of a &field; on a &mesh;.</para>'
    )
FieldCompOutput.connect("field", FieldOutput)
FieldCompOutput.aliasParam('field:field', 'field')

FieldDerivCompOutput = ComponentOutput.clone(
    name="field derivative component",
    tip='Compute a component of a Field derivative.',
    discussion='<para>Compute a single component of a &field; derivative on a &mesh;.</para>')
FieldDerivCompOutput.connect("field", FieldDerivOutput)
FieldDerivCompOutput.aliasParam('field:field',  'field')
FieldDerivCompOutput.aliasParam('field:derivative', 'derivative')

FluxCompOutput = ComponentOutput.clone(
    name="flux component",
    tip='Compute a component of a Flux.',
    discussion=
    '<para>Compute a single component of a &flux; on a &mesh;.</para>')
FluxCompOutput.connect("field", FluxOutput)
FluxCompOutput.aliasParam('field:flux', 'flux')

###############

def _invariant(mesh, elements, coords, field, invariant):
    return map(invariant, field)

InvariantOutput = output.Output(
    name="invariant",
    callback=_invariant,
    otype=FloatType,
    inputs=[outputval.OutputValParameter('field')],
    params=[invariant.InvariantParameter("invariant",
                                            tip=parameter.emptyTipString)])

FieldInvariantOutput = InvariantOutput.clone(
    name="field invariant",
    tip='Compute invariants of Fields.',
    discussion='<para>Compute invariants of a &field; on a &mesh;.</para>')
FieldInvariantOutput.connect("field", FieldOutput)
FieldInvariantOutput.aliasParam('field:field', 'field')

FluxInvariantOutput = InvariantOutput.clone(
    name="flux invariant",
    tip='Compute invariants of Fluxes.',
    discussion='<para>Compute invariants of a &flux; on a &mesh;.</para>')
FluxInvariantOutput.connect("field", FluxOutput)
FluxInvariantOutput.aliasParam('field:flux', 'flux')

FieldDerivInvariantOutput = InvariantOutput.clone(
    name="field derivative invariant",
    tip='Compute invariants of Field derivatives.',
    discussion='<para>Compute invariants of the derivative of a &field; on a &mesh;</para>')
FieldDerivInvariantOutput.connect("field", FieldDerivOutput)
FieldDerivInvariantOutput.aliasParam('field:field', 'field')
FieldDerivInvariantOutput.aliasParam('field:derivative', 'derivative')

#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#

from oof2.common import strfunction

def _functionOutput(mesh, elements, coords, f):
    ans = []
    for element, coordlist in zip(elements, coords):
        realcoords = [element.from_master(c[0], c[1]) for c in coordlist]
        ans.append([f(coord[0], coord[1]) for coord in realcoords])
    return utils.flatten1(ans)

FunctionOutput = output.Output(
    name="Function",
    callback=_functionOutput,
    otype=FloatType,
    params=[
    strfunction.XYStrFunctionParameter("f",
                                       default=strfunction.XYStrFunction('1.0'),
                                       tip="Function of x and y.")
    ]
    )

#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#

def _pointSum(mesh, elements, coords, point1, point2, a, b):
    ans = [a*f+b*s for f,s in zip(point1, point2)]
    return ans

PointSumOutput = output.Output(
    name="point sum",
    callback=_pointSum,
    otype=(coord.Coord, primitives.Point),
    inputs=[coord.CoordParameter("point1"), coord.CoordParameter("point2")],
    params=[parameter.FloatParameter("a", default=1.0),
            parameter.FloatParameter("b", default=1.0)]
    )

#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#

# Scale values to lie between 0 and 1.

def _rescaleOutput(mesh, elements, coords, minimum, maximum, inputdata):
    minval = min(inputdata)
    maxval = max(inputdata)
    def rescale(x, mn=minval, mx=maxval, tmin=minimum, tmax=maximum):
        if mx == mn:
            return 0.5*(tmin + tmax)    # arbitrary
        return tmin + (tmax-tmin)*(x-mn)/(mx-mn)
    return map(rescale, inputdata)

RescaleOutput = output.Output(
    name="rescale",
    otype=FloatType,
    callback=_rescaleOutput,
    params=[parameter.FloatParameter("minimum", default=0.),
            parameter.FloatParameter("maximum", default=1.)],
    inputs = [parameter.FloatParameter("inputdata")]
    )

#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#=-=#

# Multiply by a scalar

def _multOutput(mesh, elements, coords, values, factor, scalee):
    return [factor*x for x in scalee]

ScalarMultiplyOutput = output.Output(
    name="scalar multiply",
    callback=_multOutput,
    otype = outputval.OutputValPtr,
    params=[parameter.FloatParameter("factor", default=1.)],
    inputs=[outputval.OutputValParameter('scalee')],
    tip="Multiply the input by a constant scalar factor.")

NegateOutput = ScalarMultiplyOutput.clone(
    name="negate",
    params=dict(factor=-1),
    tip="Multiply the input by -1.")
