// -*- C++ -*-
// $RCSfile: flux.h,v $
// $Revision: 1.41 $
// $Author: reida $
// $Date: 2006/03/23 17:02:34 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#ifndef FLUX_H
#define FLUX_H

class Flux;
class FluxData;

#include "common/pythonexportable.h"
#include "common/identification.h"
#include "fieldindex.h"
#include "matvec.h"
#include "common/space.h"
#include "engine/sparselink.h"
#include "engine/linearizedsystem.h"
#include <vector>
#include <string>
#include <Python.h>

class BoundaryEdge;
class Coord;
class EdgeGaussPoint;
class EdgeNodeIterator;
class Element;
class ElementFuncNodeIterator;
class Equation;
class FluxNormal;
class Field;
class FieldIndex;
class IteratorP;
class GaussPoint;
class MasterPosition;
class FEMesh;
class OutputValue;

// The FluxData class holds the flux matrix and related values.  This
// data can't be stored in the Flux object itself, because each Flux
// is a global object, and might, in a threaded implementation, be
// doing more than one simultaneous calculation.  FluxData objects
// hold calculation-dependent parts of the Flux.  FluxData is created
// by Flux::initializeData() and should be deleted by the routine that
// called initializeData.

class FluxData {
private:
  // The fluxmatrix for an element at a point is the matrix that give
  // the flux, when multiplied by the vector of degrees of freedom.
  MATRIX_D *fluxmatrix_;
  // The "fluxoffset" contains direct contributions to the flux from
  // the properties.  It's distinguished from the similar-seeming
  // fluxrhs by the fact that the fluxrhs contains components of the
  // divergence of the flux (e.g., body forces), whereas the fluxoffset
  // contains components of the flux itself (e.g., constant fluxes, such 
  // as thermal strains times stiffness constant).  It's handled alongside
  // the fluxmatrix, and in fact really ought to be part of some kind
  // of aggregate which includes both.
  VECTOR_D *fluxoffset_;
  // The contribution of this flux's divergence to the right-hand side.
  // This vector contains body forces, not pieces of the flux itself.
  // object is analogous to the fluxmatrix, as the name suggests.
  VECTOR_D *fluxrhs_;
public:
  FluxData(const Flux*, const Element *el);
  ~FluxData();
  // These are used by the Properties to load the flux matrix.
  double matrix_element(const FEMesh*, const FieldIndex&, const Field*,
			const FieldIndex&, const ElementFuncNodeIterator&)
    const;
  double &matrix_element(const FEMesh*, const FieldIndex&, const Field*,
			const FieldIndex&, const ElementFuncNodeIterator&);
  // These versions are for use by scalar fields
  double &matrix_element(const FEMesh*,
			 const FieldIndex &fluxindex, const Field*,
			 const ElementFuncNodeIterator&);
  double matrix_element(const FEMesh*,
			const FieldIndex &fluxindex, const Field*,
			const ElementFuncNodeIterator&) const;
  
  // For offsets to the flux.
  double offset_element(const FEMesh*, const FieldIndex&) const;
  double &offset_element(const FEMesh*, const FieldIndex&);

  // Similar for rhs, to do body forces.
  double rhs_element(int) const;
  double &rhs_element(int);

  const MATRIX_D *fluxmatrix() const { return fluxmatrix_; }
  const VECTOR_D *fluxrhs() const { return fluxrhs_; }
  const VECTOR_D *fluxoffset() const { return fluxoffset_; }
};

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

class Flux : public IdentifiedObject, public PythonExportable<Flux> {
private:
  static const std::string modulename_;
  const std::string name_;
  int index_;
protected:
  // Eqnlist made protected so that the boundary_integral methods of
  // the derived classes can see it.
  std::vector<Equation*> eqnlist;

  int dim;			// number of components
  int divdim;			// number of components in divergence
public:
  Flux(const std::string &name, int dimension, int divdim);
  virtual ~Flux() {}
  static std::vector<Flux*> &allfluxes();
  static Flux *getFlux(const std::string &name);
  int index() const { return index_; }
  virtual const std::string &classname() const = 0;
  virtual const std::string &modulename() const { return modulename_; }

  int ndof() const { return dim; }
  virtual int divergence_dim() const { return divdim; }

  void addEquation(Equation*);	// this flux appears in this equation

  const std::string &name() const { return name_; }


  virtual IteratorP iterator(Planarity) const = 0;
  virtual IteratorP divergence_iterator() const = 0;
  virtual IteratorP out_of_plane_iterator() const = 0;
  virtual IndexP componenttype() const = 0;
  virtual IndexP getIndex(const std::string&) const = 0;
  virtual IndexP divergence_componenttype() const = 0;
  virtual IndexP divergence_getIndex(const std::string&) const = 0;

  virtual FluxNormal *BCCallback(const Coord&,
				 double, 
				 const Coord&,
				 const double, const double, 
				 PyObject *, const PyObject *) 
    const = 0; 

  // Create a FluxData object for use while accumulating the
  // components of the local stiffness matrix, flux offsets, and right
  // hand side.  This returns a 'new'd object.
  FluxData *initializeData(const Element *) const;

  virtual VECTOR_D *contract(const FEMesh*, const Element*, 
			     const EdgeGaussPoint&) const = 0;

  VECTOR_D *evaluate(const FEMesh*, const Element*, const MasterPosition&)
    const;
  // Non-pointer version of evaluate -- couldn't override due to SWIG.
  VECTOR_D evaluate_flux(const FEMesh*, const Element*, const MasterPosition&)
    const;
  OutputValue output(const FEMesh*, const Element*, const MasterPosition&)
    const;
  virtual OutputValue newOutputValue() const = 0;

  int eqn_integration_order(const FEMesh*, const Element*) const;

  void make_stiffness(FEMesh*, const Element*, const GaussPoint&,
		      const FluxData*, LinearizedSystem &);

  // New, possibly temporary, routine to do the RHS.
  // analogous to make_stiffness.
  void make_rhs(FEMesh*, const Element*, const GaussPoint&, 
		const FluxData*, LinearizedSystem &);

  void boundary_integral(FEMesh*, const BoundaryEdge*, const EdgeGaussPoint&,
			 const FluxNormal *) const;

  // assuming that the flux matrix has been calculated, compute the
  // matrix which, when integrated over the element and multiplied by
  // the vector of dofs, gives the divergence of the flux (up to
  // boundary terms which come from integration by parts).  Actually,
  // it's only the weak form of the divergence: the divergence
  // multiplied by the given shape function and integrated over the
  // element.  This is the local stiffness matrix for divergence
  // equations.  The minus sign from the integration by parts must be
  // included here.

  virtual void local_stiffness(const Element*, const ElementFuncNodeIterator&,
			       const GaussPoint&, const FluxData*, MATRIX_D&)
    const = 0;

  // Takes offsets to the flux stored in the flux_offset vector, 
  // and computes their contribution to the divergence, putting them
  // in the rhs.  The equation provides a local variable to store these
  // in, so this can be const.
  virtual void make_offset(const Element*, const ElementFuncNodeIterator&,
			   const GaussPoint&, const FluxData*, VECTOR_D&)
    const = 0;

  // Routine that calls the NeumannBCApp's integrate function with the
  // correct callback.
  virtual void local_boundary(const BoundaryEdge*, EdgeNodeIterator&, 
			      const EdgeGaussPoint&, const FluxNormal *,
			      std::vector<double>&) const = 0;

  friend bool operator==(const Flux&, const Flux&);
  friend bool operator!=(const Flux&, const Flux&);
  friend class FluxData;
};

std::ostream &operator<<(std::ostream&, const Flux&);

// // for std::pair
// inline bool operator<(const Flux &a, const Flux &b) {
//   return a.name()!=b.name();
// }

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

#define VEC_FLUX_DIM 3
#define VEC_DIV_DIM 1	// dimension of the divergence

class VectorFlux : public Flux {
public:
  VectorFlux(const std::string &name)
    : Flux(name, VEC_FLUX_DIM, VEC_DIV_DIM)
  {}

  virtual ~VectorFlux() {}
  virtual const std::string &classname() const;

  virtual void local_stiffness(const Element*, const ElementFuncNodeIterator&,
			       const GaussPoint&, const FluxData*, MATRIX_D&)
    const;
  virtual void local_boundary(const BoundaryEdge*, EdgeNodeIterator&, 
			      const EdgeGaussPoint&, 
			      const FluxNormal *,
			      std::vector<double>&) const;
  virtual void make_offset(const Element*, const ElementFuncNodeIterator&,
			   const GaussPoint&, const FluxData*, VECTOR_D&) const;

  virtual IteratorP iterator(Planarity) const;
  virtual IteratorP divergence_iterator() const;
  virtual IteratorP out_of_plane_iterator() const;
  virtual IndexP componenttype() const;
  virtual IndexP getIndex(const std::string&) const;
  virtual IndexP divergence_componenttype() const;
  virtual IndexP divergence_getIndex(const std::string&) const;

  virtual VECTOR_D *contract(const FEMesh*, const Element*,
			     const EdgeGaussPoint&) const;
  virtual OutputValue newOutputValue() const;

  virtual FluxNormal *BCCallback(const Coord&, 
				 double, 
				 const Coord&,
				 const double, const double, 
				 PyObject *, const PyObject *) 
    const;
};

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

#define SYMTEN_FLUX_DIM 6	// independent components of a 3x3 symm. matrix
#define SYMTEN_DIV_DIM spacedim // dimension of the divergence

// The divergence f_z = d_i sigma_iz is *two* dimensional when
// spacedim is 2, even though the flux sigma is always 3x3 tensor.  If
// the flux is in-plane, then f_z = 0 since sigma_iz = 0.  If the flux
// isn't in-plane, then f_z can be solved for after sigma_iz is found,
// but it can't be specified as part of the problem.

class SymmetricTensorFlux : public Flux {
public:
  SymmetricTensorFlux(const std::string &name)
    : Flux(name, SYMTEN_FLUX_DIM, SYMTEN_DIV_DIM)
  {}
  virtual ~SymmetricTensorFlux() {}
  virtual const std::string &classname() const;

  virtual void local_stiffness(const Element*, const ElementFuncNodeIterator&,
			       const GaussPoint&, const FluxData*, MATRIX_D&)
    const;
  virtual void local_boundary(const BoundaryEdge*, EdgeNodeIterator&,
			      const EdgeGaussPoint&, 
			      const FluxNormal *,
			      std::vector<double>&) const;
  virtual void make_offset(const Element*, const ElementFuncNodeIterator&,
			   const GaussPoint&, const FluxData*, VECTOR_D&) const;

  virtual VECTOR_D *contract(const FEMesh*, const Element*,
			     const EdgeGaussPoint&) const;

  virtual OutputValue newOutputValue() const;

  virtual IteratorP iterator(Planarity) const;
  virtual IteratorP divergence_iterator() const;
  virtual IteratorP out_of_plane_iterator() const;
  virtual IndexP componenttype() const;
  virtual IndexP getIndex(const std::string&) const;
  virtual IndexP divergence_componenttype() const;
  virtual IndexP divergence_getIndex(const std::string&) const;

  // Ideally, all arguments here would be const, but the first
  // PyObject* gets passed to PyEval_CallObject, which itself does
  // not have const arguments.
  virtual FluxNormal *BCCallback(const Coord&,
				 double,
				 const Coord&,
				 const double, const double, 
				 PyObject *, const PyObject *)
    const;
};


// Utilities used by the Python interface
Flux *getFluxByIndex(int);
int countFluxes();

#endif
