// -*- C++ -*-
// $RCSfile: linearizedsystem.h,v $
// $Revision: 1.4 $
// $Author: langer $
// $Date: 2006/06/26 21:32:06 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#ifndef LINEARIZED_SYSTEM_H
#define LINEARIZED_SYSTEM_H

#include <vector>
#include <iostream>
#include "sparselink.h"
#include "engine/matvec.h"

class CSRmatrix;
class FEMesh;

// Data object to hold the three matrices which make up the general
// time-dependent FE problem.  We call these the "stiffness",
// "kinetic", and "dynamic" matrices.  The stiffness matrix multiplies
// the degrees of freedom directly -- it's the usual stiffness matrix.
// The kinetic matrix is the matrix which multiplies the first time
// derivatives of the degrees of freedom, and the dynamic matrix
// multiplies the second time derivatives.  There are also two
// right-hand-sides, one for the contribution from body forces, and
// another for the integration over Neumann boundary conditions.
class LinearizedSystem {
private:
  FEMesh *mesh;
  SparseLinkMatrix<double> *stiffness, *kinetic, *dynamic;
  VECTOR_D *body_rhs, *boundary_rhs;
  int rows, cols; 
  bool initialized;
  std::vector<int> subgroup(const std::vector<int> group, 
			    const std::vector<int> &map) const;
  bool has_kinetics, has_dynamics;
public:
  LinearizedSystem(FEMesh *);

  
  ~LinearizedSystem();

  void initialize(int nr, int nc);
  bool is_initialized() const { return initialized; }

  void set_kinetic();
  void set_dynamic();


  bool is_kinetic() const { return has_kinetics; }
  bool is_dynamic() const { return has_dynamics; }

  void clear();

  // The rowgroup and colgroup vectors are filled in during the
  // stiffness-matrix construction phase (by equation.C).  For each
  // row or column of the matrix, the corresponding vector value is -1
  // if the DOF or equation is in-plane, and otherwise is equal to the
  // node number.  This info is used by the block preconditioner to
  // group matrix elements by nodes.
  std::vector<int> rowgroup, colgroup;

  int nrows() { return rows; }
  int ncols() { return cols; }
  bool is_nonempty_row(int&) const;
  bool is_nonempty_col(int&) const;
  
  // NB order of arguments is important here, getting them
  // backwards is wrong but will not generate an error. 

  CSRmatrix stiffness_submatrix(const std::vector<int> &rowmap,
				const std::vector<int> &colmap) const;
  CSRmatrix kinetic_submatrix(const std::vector<int> &rowmap,
			      const std::vector<int> &colmap) const;
  CSRmatrix dynamic_submatrix(const std::vector<int> &rowmap,
			      const std::vector<int> &colmap) const;
  // These functions use the appropriate maps to make sublists of the
  // rowgroup and colgroup vectors analogous to the way the submatrix
  // functions work.

  // Retrieval functions for the contained matrices. 
  SparseLinkMatrix<double> &get_stiffness() const { return *stiffness; }
  SparseLinkMatrix<double> &get_kinetic()  const { return *kinetic; }
  SparseLinkMatrix<double> &get_dynamic() const { return *dynamic; }

  VECTOR_D &get_body_rhs() const { return *body_rhs; }
  VECTOR_D &get_boundary_rhs() const { return *boundary_rhs; }

  void clear_brhs();

  // These functions use the appropriate maps to make sublists of the
  // rowgroup and colgroup vectors analogous to the way the submatrix
  // functions work.
  std::vector<int> rowgroups(const std::vector<int> &rowmap) const;
  std::vector<int> colgroups(const std::vector<int> &colmap) const;
};

#endif // LINEARIZED_SYSTEM_H
