// -*- C++ -*-
// $RCSfile: stressfreestrain.C,v $
// $Revision: 1.2 $
// $Author: langer $
// $Date: 2006/07/12 15:04:19 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "engine/IO/propertyoutput.h"
#include "engine/cstrain.h"
#include "engine/element.h"
#include "engine/flux.h"
#include "engine/material.h"
#include "engine/property/elasticity/cijkl.h"
#include "engine/property/elasticity/elasticity.h"
#include "engine/property/orientation/orientation.h"
#include "engine/property/stressfreestrain/stressfreestrain.h"

StressFreeStrain::StressFreeStrain(PyObject *reg,
			 const std::string &nm)
  : Property(nm, reg)
{
  stress_flux = dynamic_cast<SymmetricTensorFlux*>(Flux::getFlux("Stress"));
}

int StressFreeStrain::integration_order(const FEMesh*, const Element*el) const {
  return el->shapefun_degree();
}

void StressFreeStrain::fluxrhs(const FEMesh *mesh, const Element *element,
			  const Flux *flux, FluxData *fluxdata,
			  const MasterPosition &x)
  const
{
  if(*flux != *stress_flux) {
    throw ErrProgrammingError("Unexpected flux.", __FILE__, __LINE__);
  }
  const Cijkl modulus = elasticity->cijkl(mesh, element, x);

  for(SymTensorIterator ij; !ij.end(); ++ij) {
    double &offset_el = fluxdata->offset_element(mesh, ij);
    for(SymTensorIterator kl; !kl.end(); ++kl) {
      if(kl.diagonal()) {
	offset_el -= modulus(ij, kl)*stressfreestrain[kl];
      }
      else {
	offset_el -= 2.0*modulus(ij, kl)*stressfreestrain[kl];
      }
    }
  }
}

IsotropicStressFreeStrain::IsotropicStressFreeStrain(PyObject *registry,
						     const std::string &name,
						     double e)
  : StressFreeStrain(registry, name),
    e_(e)
{}

void IsotropicStressFreeStrain::cross_reference(Material *mat) {
  elasticity = dynamic_cast<Elasticity*>(mat->fetchProperty("Elasticity"));
}

void IsotropicStressFreeStrain::precompute() {
  StressFreeStrain::precompute();
  stressfreestrain(0,0) = stressfreestrain(1,1) = stressfreestrain(2,2) = e_;
}

AnisotropicStressFreeStrain::AnisotropicStressFreeStrain(PyObject *registry, 
					       const std::string &name,
					       SymmMatrix3 *e)
  : StressFreeStrain(registry, name),
    e_(*e)
{}

void AnisotropicStressFreeStrain::cross_reference(Material *mat) {
  elasticity = dynamic_cast<Elasticity*>(mat->fetchProperty("Elasticity"));
  orientation = dynamic_cast<OrientationProp*>
    (mat->fetchProperty("Orientation"));
}

void AnisotropicStressFreeStrain::precompute() {
  stressfreestrain = e_.transform(orientation->eulerangle());
}

void StressFreeStrain::output(const FEMesh *mesh,
			      const Element *element,
			      const PropertyOutput *output,
			      const MasterPosition &pos,
			      OutputVal *data)
  const
{
  // Mostly copied from ThermalExpansion::output
  const std::string &outputname = output->name();
  if(outputname == "Strain") {
    std::string stype = output->getRegisteredParamName("type");
    SymmMatrix3 *sdata = dynamic_cast<SymmMatrix3*>(data);
    if(stype == "Elastic") 
      *sdata -= stressfreestrain; 
  }
  if(outputname == "Energy") {
    std::string etype = output->getEnumParam("etype");
    if(etype == "Total" || etype == "Elastic") {
      ScalarOutputVal *edata = dynamic_cast<ScalarOutputVal*>(data);
      const Cijkl modulus = elasticity->cijkl(mesh, element, pos);
      SymmMatrix3 strain;
      findGeometricStrain(mesh, element, pos, strain);
      SymmMatrix3 stress(modulus*stressfreestrain);
      double e = 0;
      for(int i=0; i<3; i++) {
	e += stress(i,i)*strain(i,i);
	int j=(i+1)%3;
	e += 2*stress(i,j)*strain(i,j);
      }
      *edata += -0.5*e;
    }
  }
}
