# -*- python -*-
# $RCSfile: refine.py,v $
# $Revision: 1.98 $
# $Author: reida $
# $Date: 2006/07/26 18:05:58 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


# Refinement works in two stages.  First a RefinementTarget object (from
# refinementtarget.py) checks the elements and marks edges for
# subdivision.  Then a RefinementRuleSet (from refinemethod.py) is
# used to actually refine the elements, according to how many times
# each edge is marked.

from oof2.SWIG.common.IO import progressbar
from oof2.common import debug
from oof2.common import enum
from oof2.common import parallel_enable
from oof2.common import registeredclass
from oof2.common.IO import parameter
from oof2.common.IO import xmlmenudump
from oof2.engine import refinementtarget
from oof2.engine import refinemethod
from oof2.engine import refinequadbisection
from oof2.engine import skeletonmodifier
from oof2.engine import skeletonsegment
from oof2.engine import skeletonnode

SkeletonSegment = skeletonsegment.SkeletonSegment

arbitrary_factor = 128          # used by findSignature().  Must be
				# larger than the largest number of
				# refinements of an edge.

################################

# The RefinementDegree classes specify how many times edges are
# divided initially by the RefinementTarget object, and also provide a
# 'markExtras' routine that marks additional edges for division after
# the check and mark stage.

# Changes as of Mar 04.
# Refinement Degree:
# - TriSection
# - BiSection
# Each degree has two rule sets.
# - Conservative: preserves topology, that is, if you refine a quad, you'll
#                 get only quads.
# - Liberal: MAY break topology

class RefinementDegree(registeredclass.RegisteredClass):
    registry = []
    def markExtras(self, skeleton, markedEdges):
        pass

    tip = "Number of subdivisions per segment."
    discussion = xmlmenudump.loadFile('DISCUSSIONS/engine/reg/refinementdegree.xml')

class Bisection(RefinementDegree):
    divisions = 1
    def __init__(self, rule_set):
        self.rule_set = rule_set
    def markExtras(self, skeleton, markedEdges):
        if self.rule_set.name == "conservative":
            refinequadbisection.markExtras(skeleton, markedEdges)
        
registeredclass.Registration(
    'Bisection',
    RefinementDegree,
    Bisection,
    1,
    params = [enum.EnumParameter('rule_set', refinemethod.RuleSet,
                                 refinemethod.conservativeRuleSetEnum(),
                                 tip='How to subdivide elements.')
    ],
    tip="Divide element edges into two.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/bisection.xml'))

class Trisection(RefinementDegree):
    divisions = 2
    def __init__(self, rule_set):
        self.rule_set = rule_set

registeredclass.Registration(
    'Trisection',
    RefinementDegree,
    Trisection,
    0,
    params = [enum.EnumParameter('rule_set', refinemethod.RuleSet,
                                 refinemethod.conservativeRuleSetEnum(),
                                 tip='How to subdivide elements')
              ],
    tip="Gallia est omnis divisa in partes tres, as are the edges of the elements.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/trisection.xml'))

#################################

class EdgeMarkings:
    # Class for storing and returning how many divisions should be
    # performed on an edge.  Edges are defined by a pair of nodes.
    def __init__(self):
        self.markings = {}
        
    def mark(self, node0, node1, ndivs):
        # arguments are the nodes defining the edge, and the number of
        # new nodes to add to that edge.
        key = skeletonnode.canonical_order(node0, node1)
        try:
            # Only mark an edge if it's not already marked for more
            # divisions.
            if self.markings[key] < ndivs:
                self.markings[key] = ndivs
        except KeyError:
            self.markings[key] = ndivs

    def getMark(self, node0, node1):
        key = skeletonnode.canonical_order(node0, node1)
        try:
            return self.markings[key]
        except KeyError:
            return 0

    def getMarks(self, element):
        nnodes = element.nnodes()
        return [self.getMark(element.nodes[i], element.nodes[(i+1)%nnodes])
                for i in range(nnodes)]

    def getNMarkedEdges(self, element):
        marks = self.getMarks(element)
        nmarks = 0
        for m in marks:
            if m: nmarks += 1
        return nmarks

    # parallel function
    if parallel_enable.enabled():
        def getSharedSegments(self):
            pass  # Look in "refineParallel.py"

class RefinementCriterion(registeredclass.RegisteredClass):
    registry = []

    tip = "Restrict the set of Elements considered for refinement."
    discussion = """<para>

    Objects in the <classname>RefinementCriterion</classname> class
    are used as values of the <varname>criterion</varname> parameter
    when <link linkend='RegisteredClass:Refine'>refining</link>
    &skels;.  Only &elems; that satisfy the criterion are considered
    for refinement.

    </para>"""


class Unconditionally(RefinementCriterion):
    def __call__(self, skeleton, element):
        return 1
registeredclass.Registration(
    'Unconditional',
    RefinementCriterion,
    Unconditionally,
    ordering=0,
    tip='Consider all Elements for possible refinement.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/unconditionally.xml')
    )

class AreaUnits(enum.EnumClass(
    ('Pixel', 'Area is in units of the pixel area.'),
    ('Physical', 'Area is in physical units.'))):
    tip = "Units of area."
    discussion = """<para>
    Units of element area.  See <xref
    linkend='RegisteredClass:MinimumArea'/> for details.
    </para>"""

class MinimumArea(RefinementCriterion):
    def __init__(self, threshold, units):
        self.threshold = threshold
        self.units = units
    def __call__(self, skeleton, element):
        if self.units == 'Pixel':
            return element.area() > self.threshold*skeleton.MS.areaOfPixels()
        else:
            return element.area() > self.threshold

registeredclass.Registration(
    'Minimum Area',
    RefinementCriterion,
    MinimumArea,
    ordering=1,
    params=[parameter.FloatParameter('threshold', 10,
                                     tip="Minimum acceptable element area."),
            enum.EnumParameter('units', AreaUnits, AreaUnits('Pixel'),
                               tip='Units for the minimum area')],
    tip='Only refine elements with area greater than the given threshold.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/minimumarea.xml')
    )

###################################
            
class Refine(skeletonmodifier.SkeletonModifier):
    def __init__(self, targets, criterion, degree, alpha):
        self.targets = targets      # RefinementTarget instance
        self.criterion = criterion  # Criterion for refinement
        self.degree = degree        # RefinementDegree instance
        # alpha is used for deciding between different possible
        # refinements, using effective energy
        self.alpha = alpha 
        # Available rules for the specified rule_set of refinement degree
        self.rules = refinemethod.getRuleSet(self.degree.rule_set.name)
    #########

    def applyAMR(self, skeleton, context): # for adaptive mesh refinement
        newSkeleton = skeleton.improperCopy(fresh=True)
        return self.refinement(skeleton, newSkeleton, context)
    
    def apply(self, skeleton, context):
        newSkeleton = skeleton.improperCopy()  # fresh=False
        return self.refinement(skeleton, newSkeleton, context)

    def refinement(self, skeleton, newSkeleton, context):
        # pBar.set_message("marking elements...       ")
        pBar = progressbar.getProgress()
        # Copy the old skeleton, without copying the elements.
##         newSkeleton = skeleton.improperCopy()
        markedEdges = EdgeMarkings()
        self.newEdgeNodes = {}          # allows sharing of new edge nodes

        # Primary marking
        self.targets(skeleton, context, self.degree.divisions, markedEdges,
                     self.criterion)
        # Additional marking
        self.degree.markExtras(skeleton, markedEdges)

        # Refine elements and segments
        segmentdict = {}                # which segments have been handled
        elements = skeleton.elements
        n = len(elements)
        
        for ii in range(n):
            oldElement = elements[ii]
            oldnnodes = oldElement.nnodes()
            # Get list of number of subdivisions on each edge ("marks")
            marks = markedEdges.getMarks(oldElement)
            # Find the canonical order for the marks. (The order is
            # ambiguous due to the arbitrary choice of the starting
            # edge.  Finding the canonical order allows the refinement
            # rule to be found in the rule table.)  rotation is the
            # offset into the elements node list required to match the
            # refinement rule to the element's marked edges.
            # signature is the canonical ordering of the marks.
            rotation, signature = findSignature(marks)
            # Create new nodes along the subdivided element edges
            edgenodes = [self.getNewEdgeNodes(oldElement.nodes[i],
                                              oldElement.nodes[(i+1)%oldnnodes],
                                              marks[i], newSkeleton)
                         for i in range(oldnnodes)]
            # Create new elements
            newElements = self.rules[signature].apply(oldElement, rotation,
                                                      edgenodes, newSkeleton,
                                                      self.alpha)
            # If the old element's homogeneity is "1", it's safe to say that
            # new elements' homogeneities are "1".
            if oldElement.homogeneity(skeleton) == 1.:
                for el in newElements:
                    el.setHomogeneityValues(oldElement.cachedHomogData.value)

            # The calls to Skeleton.newElement() made by the
            # refinement rules have created new SkeletonSegments in
            # newSkeleton, but have not set the parentage of those
            # segments.  We have to fix that here.
            for newElement in newElements:
                for segment in newElement.getSegments(newSkeleton):
                    # Only look at each segment once.
                    if not segmentdict.has_key(segment):
                        segmentdict[segment] = 1
                        pseg = findParentSegment(skeleton, newElement,
                                                 segment,
                                                 edgenodes)
                        if pseg:
                            pseg.add_child(segment)
                            segment.add_parent(pseg)
            if pBar.query_stop():
                pBar.set_failure()
                pBar.set_message("Failed")
                return None
            else:
                pBar.set_progress(1.0*(ii+1)/n)
                pBar.set_message("refining skeleton: %d/%d" % (ii+1, n))

            
        newSkeleton.cleanUp()
        return newSkeleton

    ################
    
    def getNewEdgeNodes(self, node0, node1, ndivs, newSkeleton):
        # Create ndivs new nodes between node0 and node1 in the
        # skeleton.
        if ndivs < 1:
            return []
        key = skeletonnode.canonical_order(node0, node1)
        try:
            nodes = self.newEdgeNodes[key]
            # Since this is the second time we're using this list of
            # nodes, we must be looking at them from the other side,
            # and the nodes should be in the opposite order.
            # Reversing them in place like this would be wrong if the
            # list weren't being used immediately, as it is in
            # Refine.apply()
            nodes.reverse()
        except KeyError:
            nodes = [None]*ndivs
            p0 = node0.position()
            p1 = node1.position()
            delta = (p1 - p0)/(ndivs+1)
            for i in range(ndivs):
                pt = p0 + (i+1)*delta
                newnode = newSkeleton.newNode(pt.x, pt.y)
                nodes[i] = newnode
            self.newEdgeNodes[key] = nodes
        return nodes

    
    # parallel fuctions
    if parallel_enable.enabled():
        def apply_parallel(self, skeleton, context):
            newSkeleton = skeleton.improperCopy()  # fresh=False
            return self.refinement_parallel(skeleton, newSkeleton, context)

        def refinement_parallel(self, skeleton, newSkeleton, context):
            pass  # Look in "refineParallel.py"

        def shareCommonNodes(self, markedEdges, skeleton, newSkeleton):
            pass  # Look in "refineParallel.py"
    
#################################

def findParentSegment(oldSkeleton, element, segment, edgenodes):
    n0, n1 = segment.nodes()            # nodes of the segment
    parents = n0.getParents()
    if parents:
        p0 = parents[-1]
    else:
        p0 = None
        
    parents = n1.getParents()
    if parents:
        p1 = parents[-1]
    else:
        p1 = None    
    oldElement = element.getParents()[-1]
    nnodes = oldElement.nnodes()

    if p0 is not None and p1 is not None:
        pseg = oldSkeleton.findSegment(p0, p1)
        if pseg:
            # Trivial case: the parent nodes define a segment in the
            # parent skeleton.  That segment is this segment's parent.
            return pseg
        # Both nodes have parents, but the parent nodes don't define a
        # segment in the old skeleton.  (This is unlikely to happen
        # during refinement.)  The segment does not have a parent.
        return None

    elif p0 is not None or p1 is not None:
        # Exactly one new endpoint has a parent in the old element.
        # The other endpoint may lie on an old segment.  If that
        # segment's other node is the parent of the new node, then
        # that segment is the parent of this segment.
        if p0 is None:
            parentnode = p1
            childnode = n1
            freechild = n0
        else:
            parentnode = p0
            childnode = n0
            freechild = n1
        which = oldElement.nodes.index(parentnode)
        if freechild in edgenodes[which]:
            next = (which+1)%nnodes
            return oldSkeleton.findSegment(parentnode, oldElement.nodes[next])
        prev = (which - 1 + nnodes) % nnodes
        if freechild in edgenodes[prev]:
            return oldSkeleton.findSegment(parentnode, oldElement.nodes[prev])

    elif p0 is None and p1 is None:
        # Neither new endpoint has a parent.  If the new endpoints are
        # consecutive nodes in the list of new nodes added to a parent
        # element segment, then that parent segment is the parent
        # we're looking for.
        for i in range(nnodes):
            nodelist = edgenodes[i]
            try:
                which0 = nodelist.index(n0)
                which1 = nodelist.index(n1)
            except ValueError:
                pass
            else:
                if abs(which0 - which1) == 1:
                    return oldSkeleton.findSegment(oldElement.nodes[i],
                                                   oldElement.nodes[(i+1)%
                                                                    nnodes])

################################

def findSignature(marks):
    # Given a list of subdivisions of sides (marks), rotate a
    # canonical starting point for the list, so that it can be used as
    # an index into a table of refinement functions.  For a list
    # [x,y,z], the canonical order is that which maximizes the number
    # xyz in base arbitrary_factor.  This will fail if edges are ever
    # divided into more than arbitrary_factor segments in one
    # refinement operation.
    n = len(marks)
    max = -1
    imax = None
    for i in range(n):
        key = marks[i]
        for j in range(1, n):
            key = arbitrary_factor*key + marks[(i+j)%n]
        if key > max:
            max = key
            imax = i
    return imax, tuple([marks[(i+imax)%n] for i in range(n)])


####################

registeredclass.Registration(
    'Refine',
    skeletonmodifier.SkeletonModifier,
    Refine,
    ordering=0,
    params=[parameter.RegisteredParameter('targets',
                                          refinementtarget.RefinementTarget,
                                          tip='Target elements to be refined.'),
            parameter.RegisteredParameter('criterion', RefinementCriterion,
                                          tip='Exclude certain elements.'),
            parameter.RegisteredParameter('degree', RefinementDegree,
                                    tip='Preferred way of subdividing a side.'),
            skeletonmodifier.alphaParameter
            ],
    tip="Subdivide elements.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/refine.xml'))
