// -*- C++ -*-
// $RCSfile: solver.C,v $
// $Revision: 1.48 $
// $Author: reida $
// $Date: 2006/03/23 17:02:35 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include <math.h>
#include "SparseLib++/bicg.h"
#include "SparseLib++/bicgstab.h"
#include "SparseLib++/cg.h"
#include "SparseLib++/gmres.h"
#include "common/IO/progressbar.h"
#include "common/tostring.h"
#include "engine/csrmatrix.h"
#include "engine/femesh.h"
#include "engine/matvec.h"
#include "engine/smallmatrix.h"
#include "engine/solver.h"

void LinearSolver::set_blocking(const std::vector<int> rb, 
			  const std::vector<int> cb) {
  rowblocks = rb;
  colblocks = cb;
}

void LinearSolver::set_blocking(const FEMesh * mesh) {
  
  rowblocks = mesh->A_rowblocks();
  colblocks = mesh->A_colblocks();
  // rowblocks = mesh->get_stiffness()->rowgroups(mesh->indepeqnmap);
  // colblocks = mesh->get_stiffness()->colgroups(mesh->freedofmap);
}

GMRESSolver::~GMRESSolver() {}

bool GMRESSolver::solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x) {
  // Dump state.
//   std::cerr << "GMRESSolver object state:" << std::endl;
//   std::cerr << "Rowblocks:" << std::endl;
//   std::cerr << rowblocks << std::endl;
//   std::cerr << "Colblocks:" << std::endl;
//   std::cerr << colblocks << std::endl;

//   std::cerr << "Max iterations: " << max_iterations << std::endl;
//   std::cerr << "Tolerance: " << tolerance << std::endl;
//   std::cerr << "Number of iterations:" << number_of_iterations << std::endl;
//   std::cerr << "Residual norm:" << residual_norm << std::endl;
//   std::cerr << "Krylov dimension: " << krylov_dimension << std::endl;

//   std::cerr << "Matrix: " << std::endl;
//   std::cerr << A << std::endl;
//   std::cerr << "Right-hand side: " << std::endl;
//   std::cerr << rhs << std::endl;


  //assume preconditioner has already been selected
  //std::cerr << "GMRESSolver::solve: getting real preconditioner" << std::endl;
  PreconditionerCore * preconditioner = 
    pc->create_preconditioner(A, rowblocks, colblocks);
  //std::cerr << "GMRESSolver::solve: got preconditioner" << std::endl;
  MATRIX_D H(krylov_dimension+1, krylov_dimension);
  int krlv_dim = krylov_dimension;//will be modified by GMRES
  number_of_iterations = max_iterations;
  residual_norm = tolerance;
  status_ = GMRES(A, x, rhs, *preconditioner, H, krlv_dim,
		  number_of_iterations, residual_norm);
//   std::cerr << "Answer:" << std::endl;
//   std::cerr << x << std::endl;

  SolverMessager messenger(status_, number_of_iterations, residual_norm);
  _message = messenger._message();
  
  delete preconditioner;
  return !status_;
  
}

std::string GMRESSolver::message() const {
  return "GMRES" + _message;
}

bool BiCGSolver::solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x) {
  //assume preconditioner has already been selected
  PreconditionerCore * preconditioner  = pc->create_preconditioner
    (A, rowblocks, colblocks);
  number_of_iterations = max_iterations;//will be modifed by method
  residual_norm = tolerance;
  status_ = BiCG(A, x, rhs, *preconditioner, number_of_iterations, residual_norm);

  SolverMessager messenger(status_, number_of_iterations, residual_norm);
  _message = messenger._message();

  delete preconditioner;
  return !status_;
  
}


BiCGSolver::~BiCGSolver() {}

std::string BiCGSolver::message() const {
  return "BiCG" + _message;
}



bool CGSolver::solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x) {
  //assume preconditioner has already been selected
  PreconditionerCore * preconditioner  = pc->create_preconditioner
    (A, rowblocks, colblocks);
  residual_norm= tolerance;//will be modifed by method
  number_of_iterations = max_iterations;
  status_ = CG(A, x, rhs, *preconditioner, number_of_iterations, residual_norm);

  SolverMessager messenger(status_, number_of_iterations, residual_norm);
  _message = messenger._message();

  delete preconditioner;
  return !status_;
}


CGSolver::~CGSolver() {}

std::string CGSolver::message() const {
  return "CG" + _message;
}


bool BiCGStabSolver::solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x) {
  //assume preconditioner has already been selected
  PreconditionerCore * preconditioner  = pc->create_preconditioner
    (A, rowblocks, colblocks);
  residual_norm = tolerance;//will be modifed by method
  number_of_iterations = max_iterations;
  status_ = BiCGSTAB(A, x, rhs, *preconditioner, number_of_iterations, residual_norm);

  SolverMessager messenger(status_, number_of_iterations, residual_norm);
  _message = messenger._message();

  delete preconditioner;
  return !status_;
}


BiCGStabSolver::~BiCGStabSolver(){
  //std::cerr <<"destroyed BiCGStab solver " << std::endl;
}

std::string BiCGStabSolver::message() const {
  return "BiCG Stabilized" + _message;
}


DirectSolver::~DirectSolver() {}

bool DirectSolver::solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x) {

  SmallMatrix small_m(A.nrows(), A.ncols());
  small_m.clear();

  Progress *cpb = getProgress();

  for(int i=0; i<A.nrows() && !cpb->query_stop(); i++) 
    for(int j=0; j<A.ncols() && !cpb->query_stop(); j++)
      small_m(i,j) = A(i,j);  

  SmallMatrix small_r(rhs.size(),1);
  for(VECTOR_D::size_type i=0; i<rhs.size() && !cpb->query_stop(); i++)
    small_r(i,0) = rhs(i);

  int rcode = small_m.solve(small_r);

  double r=0.0;
  if(rcode==0) {
    for(VECTOR_D::size_type i=0; i<rhs.size() && !cpb->query_stop(); i++)
      x[i] = small_r(i,0);
    status_ = true;
    // cpb->set_success();
    
    // Compute square of residual.
    double *resid = new double[A.nrows()];
    for(int i=0;i<A.nrows() && !cpb->query_stop();i++) {
      resid[i]=rhs(i);
      for(int j=0;j<A.ncols() && !cpb->query_stop();j++) 
	resid[i]-=A(i,j)*x[j];
      r += resid[i]*resid[i];
    }
    delete [] resid;
    
  }
  else {	
    status_ = false;
    cpb->set_failure();
  }
  if (cpb->query_stop())
    cpb->set_failure();


  SolverMessager messenger((status_==true?0:1), 1, sqrt(r));
  _message = messenger._message();

  return status_;
}


std::string DirectSolver::message() const {
  return "Direct" + _message;
}



// Why does this class exist?  The way it's used, it would be easier
// if it were just a function.
SolverMessager::SolverMessager(const int state, const int num_iter,
			       const double res)
{
  switch (state) {
    case 0:
      message = " converged!\n";
      break;
    case 1:
      message = " FAILED to converge!\n";
      break;
    default:
      message = " Unknown state";
      break;
  };
  std::string solver_info = "residual = " + to_string(res) + " number of completed iterations = " + to_string(num_iter);
  message += solver_info;

}

// The message is also the "repr" -- this can include some state data,
// and is potentially confusing, but it's mainly for type-identifying
// and debugging, so hopefully it won't confuse actual users.
// Function uses only public data, and so is nobody's friend.
std::ostream & operator<<(std::ostream &os, const LinearSolver &s) {
  os << s.message();
  return os;
}

