// -*- C++ -*-
// $RCSfile: sparselink.h,v $
// $Revision: 1.10 $
// $Author: langer $
// $Date: 2006/02/16 21:07:21 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */


// Sparse row matrix stored with linked lists, for easy construction.
// Meant to be used as temporary storage while constructing compressed
// row matrix.

#ifndef SPARSELINK_H
#define SPARSELINK_H

#include <iostream>

template <class TYPE> class SLMentry;
template <class TYPE> class SparseLinkMatrix;
template <class TYPE> class SparseLinkIterator;
template <class TYPE> class CSparseLinkIterator;
template <class TYPE> std::ostream &operator<<(std::ostream&,
 					       const SparseLinkMatrix<TYPE>&);

#include "engine/ooferror.h"
#include <vector>
#include <math.h>
#include <string>


template <class TYPE>
class SLMentry {
private:
  SLMentry(int c) : next(0), val(0), col(c) {}
  SLMentry(int c, TYPE v) : next(0), val(v), col(c) {}
  SLMentry *next;
  TYPE val;
  int col;
  friend std::ostream &operator<< <>(std::ostream&,
 				     const SparseLinkMatrix<TYPE>&);
  friend class SparseLinkMatrix<TYPE>;
  friend class SparseLinkIterator<TYPE>;
  friend class CSparseLinkIterator<TYPE>;
};

template <class TYPE>
class SparseLinkMatrix {
private:
  int numrows, numcols, nz;
  int stretchy;	// resized on the fly? no bounds check
  std::vector<SLMentry<TYPE>*> row;
  const TYPE zero; // a const reference to this is returned by const
		// access functions
  std::vector<bool> nonempty_row;
  std::vector<bool> nonempty_col;
public:
  SparseLinkMatrix();
  SparseLinkMatrix(int, int);
  SparseLinkMatrix(const SparseLinkMatrix<TYPE>&);
  virtual ~SparseLinkMatrix();

  typedef SparseLinkIterator<TYPE> iterator;
  typedef CSparseLinkIterator<TYPE> const_iterator;
  CSparseLinkIterator<TYPE> begin() const;
  CSparseLinkIterator<TYPE> end() const;

  CSparseLinkIterator<TYPE> begin(int) const; // for iterating over a row
  CSparseLinkIterator<TYPE> end(int) const; // for iterating over a row

  void clear(void);	// remove all entries
  void setsize(int, int);	// set size and make nonstretchy
  // return an entry, creating space if necessary
  TYPE &operator()(int, int);
  const TYPE &operator()(int, int) const; // don't create space
  int nrows() const { return numrows; }
  int ncols() const { return numcols; }
  int nnz() const { return nz; } // number of non-zero entries
  bool symmetric(double tol=0.0) const;
  bool is_nonempty_row(int i) const { return nonempty_row[i]; }
  bool is_nonempty_col(int j) const { return nonempty_col[j]; }

  friend std::ostream &operator<< <>(std::ostream&,
 				     const SparseLinkMatrix<TYPE>&);
  friend class SparseLinkIterator<TYPE>;
  friend class CSparseLinkIterator<TYPE>;
};

template <class TYPE>
class CSparseLinkIterator {	// not quite an STL iterator
protected:
  int row_;
  const SLMentry<TYPE> *entry;
  const SparseLinkMatrix<TYPE> *mat;
public:
  CSparseLinkIterator(const SparseLinkMatrix<TYPE> *m, int r,
		      const SLMentry<TYPE> *e)
    : row_(r), entry(e), mat(m)
  {}
  virtual ~CSparseLinkIterator() {}
  bool operator==(const CSparseLinkIterator<TYPE> &o) const;
  bool operator!=(const CSparseLinkIterator<TYPE> &o) const;
  bool operator<(const CSparseLinkIterator<TYPE> &o) const;
  void operator++();
  void validify();
  const TYPE &operator*() const { return entry->val; }
  int row() const { return row_; }
  int col() const { return entry->col; }
};

template <class TYPE>
bool CSparseLinkIterator<TYPE>::operator==(const CSparseLinkIterator<TYPE> &o)
  const
{
  if(!entry || !o.entry) return (entry == o.entry);
  return (row() == o.row() && entry->col == o.entry->col);
}

template <class TYPE>
bool CSparseLinkIterator<TYPE>::operator!=(const CSparseLinkIterator<TYPE> &o)
  const
{
  if(!entry || !o.entry) return (entry != o.entry);
  return (row() != o.row() || entry->col != o.entry->col);
}

template <class TYPE>
bool CSparseLinkIterator<TYPE>::operator<(const CSparseLinkIterator<TYPE> &o)
  const
{
  if(!entry) return false;
  if(!o.entry) return true;
  return (row() < o.row()) ||
    ((row() == o.row()) && (entry->col < o.entry->col));
}

// Special function to advance an iterator to the next valid entry, if
// there is one.
template <class TYPE>
void CSparseLinkIterator<TYPE>::validify() {
  if(entry) return;
  ++row_;
  while(row_ < mat->nrows() && !(entry = mat->row[row_]))
    ++row_;
}


template <class TYPE>
void CSparseLinkIterator<TYPE>::operator++() {
  if(entry) {
    entry = entry->next;
    if(!entry) {		// at end of a row
      ++row_;			// go to next non-empty row
      while(row_ < mat->nrows() && !(entry = mat->row[row_]))
	++row_;
    }
  }
}

template <class TYPE>
CSparseLinkIterator<TYPE> SparseLinkMatrix<TYPE>::begin() const {
  // find first non-empty row
  for(int r=0; r<numrows; r++) {
    if(row[r]) {
      return CSparseLinkIterator<TYPE>(this, r, row[r]);
    }
  }
  // didn't find any non-empty rows!
  return CSparseLinkIterator<TYPE>(this, numrows, 0);
}

template <class TYPE>
CSparseLinkIterator<TYPE> SparseLinkMatrix<TYPE>::end() const {
  return CSparseLinkIterator<TYPE>(this, numrows, 0);
}

template <class TYPE>
CSparseLinkIterator<TYPE> SparseLinkMatrix<TYPE>::begin(int r) const {
  // beginning of row r
  return CSparseLinkIterator<TYPE>(this, r, row[r]);
}

template <class TYPE>
CSparseLinkIterator<TYPE> SparseLinkMatrix<TYPE>::end(int r) const {
  // one past the end of row r
  if(r < numrows - 1) {
     CSparseLinkIterator<TYPE> iter(this, r+1, row[r+1]);
     // Make sure that the iterator is valid, or operator< won't work.
     iter.validify();
     return iter;
  }
  return CSparseLinkIterator<TYPE>(this, numrows, 0);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Template function bodies that used to be in sparselink.C are below
// here.

// fixed size
template <class TYPE>
SparseLinkMatrix<TYPE>::SparseLinkMatrix(const int n, const int m)
  : numrows(n), numcols(m), nz(0), stretchy(0), row(n, (SLMentry<TYPE>*) 0),
    zero(0), nonempty_row(n, false), nonempty_col(m, false)
{}

// variable size
template <class TYPE>
SparseLinkMatrix<TYPE>::SparseLinkMatrix()
  : numrows(0), numcols(0), nz(0), stretchy(1), row(0), zero(0),
    nonempty_row(0), nonempty_col(0)
{}

template <class TYPE>
SparseLinkMatrix<TYPE>::~SparseLinkMatrix() {
  clear();
}

template <class TYPE>
void SparseLinkMatrix<TYPE>::setsize(int n, int m) {
  clear();
  numrows = n;
  numcols = m;
  row.resize(n, 0);
  nonempty_row.resize(n, false);
  nonempty_col.resize(m, false);
  stretchy = 0;
}

// copy constructor
template <class TYPE>
SparseLinkMatrix<TYPE>::SparseLinkMatrix(const SparseLinkMatrix<TYPE> &other)
  : numrows(other.nrows()),
    numcols(other.ncols()),
    nz(0),
    stretchy(other.stretchy),
    row(other.nrows(), (SLMentry<TYPE>*) 0),
    zero(0),
    nonempty_row(other.nrows(), false),
    nonempty_col(other.ncols(), false)
{
  for(typename SparseLinkMatrix<TYPE>::const_iterator i=other.begin(); 
      i<other.end(); ++i)
    {
      (*this)(i.row(), i.col()) = *i;
      nonempty_row[i.row()] = true;
      nonempty_col[i.col()] = true;
    }
}

template <class TYPE>
void SparseLinkMatrix<TYPE>::clear() {
  for(int i=0; i<numrows; i++) {
    SLMentry<TYPE> *tmp = row[i], *next;
    while(tmp) {
      next = tmp->next;
      delete tmp;
      tmp = next;
    }
    row[i] = 0;
  }
  if(stretchy) {
    row.resize(0);
    nonempty_row.resize(0);
    nonempty_col.resize(0);
    numrows = numcols = 0;
  }
  else {
    for(std::vector<bool>::iterator i=nonempty_row.begin();
	i<nonempty_row.end(); ++i)
      *i = false;
    for(std::vector<bool>::iterator i=nonempty_col.begin();
	i<nonempty_col.end(); ++i)
      *i = false;
  }
  nz = 0;
}

// access function (i,j) creates space for element (i,j) if necessary

template <class TYPE>
TYPE &SparseLinkMatrix<TYPE>::operator()(int i, int j) {
  // look through list row(i)
#ifdef DEBUG
  if(i < 0  || j < 0 || (!stretchy && (i >= numrows || j >= numcols))) {
    std::cerr << "Array index (" << i << ", " << j
	 << ") out of bounds in SparseLinkMatrix!" << std::endl;
    //    breakpoint();
    throw
      ErrProgrammingError("Index out of bounds in SparseLinkMatrix",
			  __FILE__, __LINE__);
  }
#endif // DEBUG
  if(stretchy) {
    if(i >= numrows) {    // make space for new row
      numrows = i+1;
      row.resize(numrows, (SLMentry<TYPE>*) 0);
      nonempty_row.resize(numrows, false);
    }
    if(j >= numcols) {
      numcols = j+1;
      nonempty_col.resize(numcols, false);
    }
  }

  nonempty_row[i] = true;
  nonempty_col[j] = true;
    
  SLMentry<TYPE> *next = row[i];
  SLMentry<TYPE> *last = 0;
  SLMentry<TYPE> *newentry;
  while(next && next->col < j) {
    last = next;
    next = next->next;
  }
    
  if(!next) { // got through the whole list without finding anything
    if(last) {
      newentry = last->next = new SLMentry<TYPE>(j);
      nz++;
      return newentry->val;
    }
    // nothing in list yet
    newentry = row[i] = new SLMentry<TYPE>(j);
    nz++;
    return newentry->val;
  }
  else {	// insert before end of list
    if(next->col == j)    // entry already there
      return next->val;
	
    if(last) {  // insert between last and next
      newentry = last->next = new SLMentry<TYPE>(j);
      nz++;
      newentry->next = next;
      return newentry->val;
    }
    // insert at beginning of list
    newentry = row[i] = new SLMentry<TYPE>(j);
    nz++;
    newentry->next = next;
    return newentry->val;
  }
}
template <class TYPE>
const TYPE &SparseLinkMatrix<TYPE>::operator()(int i, int j) const {
#ifdef DEBUG
  if(i < 0  || j < 0 || i >= numrows || j >= numcols) {
    std::cerr << "Array index (" << i << ", " << j
	 << ") out of bounds in SparseLinkMatrix!" << std::endl;
    //    breakpoint();
    throw
      ErrProgrammingError("Index out of bounds in SparseLinkMatrix",
				 __FILE__, __LINE__);
  }
#endif // DEBUG
    
  SLMentry<TYPE> *next = row[i];
  while(next && next->col < j) {
    next = next->next;
  }

  if(next && next->col == j)
    return next->val;
  return zero;			// can't return a reference to 0
}

template <class TYPE>
bool SparseLinkMatrix<TYPE>::symmetric(double tol) const {
  for(typename SparseLinkMatrix<TYPE>::const_iterator i=begin(); 
      i<end(); ++i) {
    if(fabs(*i - (*this)(i.col(), i.row())) > tol)
      return false;
  }
  return true;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

template <class TYPE>
std::ostream& operator<<(std::ostream& os, const SparseLinkMatrix<TYPE>& slm) {
  for(typename SparseLinkMatrix<TYPE>::const_iterator i = slm.begin(); i<slm.end(); ++i) 
     {
       os << i.row() << " " << i.col() << " " << *i << std::endl;
     }
  return os;
}

// ------------------------------------------------- //


#endif // SPARSELINK_H
